@@ -3,6 +3,7 @@
 ELM energy wizard
 """
 from abc import ABC, abstractmethod
+from time import perf_counter
 import copy
 import os
 import json
@@ -61,8 +62,12 @@ class EnergyWizardBase(ApiBase, ABC):
             ranked strings/scores outputs.
         """
 
-    def engineer_query(self, query, token_budget=None, new_info_threshold=0.7,
-                       convo=False):
+    def engineer_query(self,
+                       query,
+                       token_budget=None,
+                       new_info_threshold=0.7,
+                       convo=False
+                       ):
         """Engineer a query for GPT using the corpus of information
 
         Parameters
@@ -79,6 +84,7 @@ class EnergyWizardBase(ApiBase, ABC):
             Flag to perform semantic search with full conversation history
             (True) or just the single query (False). Call EnergyWizard.clear()
             to reset the chat history.
+
         Returns
         -------
         message : str
@@ -87,6 +93,11 @@ class EnergyWizardBase(ApiBase, ABC):
         references : list
             The list of references (strs) used in the engineered prompt is
             returned here
+        used_index : list
+            Shows the indices of the documents used in making a query to the
+            vector database
+        vector_query_time : float
+            measures vector database query time
         """
 
         self.messages.append({"role": "user", "content": query})
@@ -99,9 +110,10 @@ class EnergyWizardBase(ApiBase, ABC):
             query = '\n\n'.join(query)
 
         token_budget = token_budget or self.token_budget
-
+        start_time = perf_counter()
         strings, _, idx = self.query_vector_db(query)
-
+        end_time = perf_counter()
+        vector_query_time = end_time - start_time
         message = copy.deepcopy(self.MODEL_INSTRUCTION)
         question = f"\n\nQuestion: {query}"
         used_index = []
@@ -125,8 +137,7 @@ class EnergyWizardBase(ApiBase, ABC):
         message = message + question
         used_index = np.array(used_index)
         references = self.make_ref_list(used_index)
-
-        return message, references, used_index
+        return message, references, used_index, vector_query_time
 
     @abstractmethod
     def make_ref_list(self, idx):
@@ -144,15 +155,16 @@ class EnergyWizardBase(ApiBase, ABC):
             ["{ref_title} ({ref_url})"]
         """
 
-    def chat(self, query,
-             debug=True,
+    def chat(self,
+             query,
              stream=True,
              temperature=0,
              convo=False,
              token_budget=None,
              new_info_threshold=0.7,
              print_references=False,
-             return_chat_obj=False):
+             return_chat_obj=False
+             ):
         """Answers a query by doing a semantic search of relevant text with
         embeddings and then sending engineered query to the LLM.
 
@@ -160,8 +172,6 @@ class EnergyWizardBase(ApiBase, ABC):
         ----------
         query : str
             Question being asked of EnergyWizard
-        debug : bool
-            Flag to return extra diagnostics on the engineered question.
         stream : bool
             Flag to print subsequent chunks of the response in a streaming
             fashion
@@ -181,26 +191,31 @@ class EnergyWizardBase(ApiBase, ABC):
             contents being added multiple times.
         print_references : bool
             Flag to print references if EnergyWizard is initialized with a
-            valid ref_col.
+            valid ``ref_col``.
         return_chat_obj : bool
-            Flag to only return the ChatCompletion from OpenAI API.
+            Flag to return the ChatCompletion object from OpenAI API instead of
+            the message string.
 
         Returns
         -------
-        response : str
-            GPT output / answer.
+        response : str | ChatCompletion
+            GPT response string ``if not return_chat_obj`` or OpenAI
+            ChatCompletion object ``if return_chat_obj``
         query : str
-            If debug is True, the engineered query asked of GPT will also be
-            returned here
+            The engineered query asked of GPT including retrieved context
         references : list
-            If debug is True, the list of references (strs) used in the
-            engineered prompt is returned here
+            The list of references (strs) used in the engineered prompt is
+            returned here
+        performance : dict | None
+            dictionary with keys of ``total_chat_time``,
+            ``chat_completion_time`` and ``vectordb_query_time``. If
+            return_chat_obj, this is None
         """
-
+        start_chat_time = perf_counter()
         out = self.engineer_query(query, token_budget=token_budget,
                                   new_info_threshold=new_info_threshold,
                                   convo=convo)
-        query, references, _ = out
+        query, references, _, vector_query_time = out
 
         messages = [{"role": "system", "content": self.MODEL_ROLE},
                     {"role": "user", "content": query}]
@@ -210,17 +225,27 @@ class EnergyWizardBase(ApiBase, ABC):
                       temperature=temperature,
                       stream=stream)
 
+        start_completion_time = perf_counter()
         response = self._client.chat.completions.create(**kwargs)
 
+        finish_completion_time = perf_counter()
+        chat_completion_time = finish_completion_time - start_completion_time
+        total_chat_time = finish_completion_time - start_chat_time
+
+        performance = {
+            "total_chat_time": total_chat_time,
+            "chat_completion_time": chat_completion_time,
+            "vectordb_query_time": vector_query_time
+        }
+
         if return_chat_obj:
-            return response, query, references
+            return response, query, references, performance
 
         if stream:
             for chunk in response:
                 chunk_msg = chunk.choices[0].delta.content or ""
                 response_message += chunk_msg
                 print(chunk_msg, end='')
-
         else:
             response_message = response.choices[0].message.content
 
@@ -235,10 +260,15 @@ class EnergyWizardBase(ApiBase, ABC):
             if stream:
                 print(ref_msg)
 
-        if debug:
-            return response_message, query, references
-        else:
-            return response_message
+        end_time = perf_counter()
+        total_chat_time = end_time - start_chat_time
+        performance = {
+            "total_chat_time": total_chat_time,
+            "chat_completion_time": chat_completion_time,
+            "vectordb_query_time": vector_query_time
+        }
+
+        return response_message, query, references, performance
 
 
 class EnergyWizard(EnergyWizardBase):
@@ -404,7 +434,7 @@ class EnergyWizardPostgres(EnergyWizardBase):
                          }
     """Optional mappings for weird azure names to tiktoken/openai names."""
 
-    DEFAULT_META_COLS = ('title', 'url', 'authors', 'year', 'category', 'id')
+    DEFAULT_META_COLS = ['title', 'url', 'nrel_id', 'id']
     """Default columns to retrieve for metadata"""
 
     def __init__(self, db_host, db_port, db_name,
@@ -431,7 +461,8 @@ class EnergyWizardPostgres(EnergyWizardBase):
             value is sqrt(n_lists).
         meta_columns : list
             List of metadata columns to retrieve from database. Default
-            query returns title and url.
+            query returns title, url, nrel_id, and id. nrel_id and id are
+            necessary to correctly format references.
         cursor : psycopg2.extensions.cursor
             PostgreSQL database cursor used to execute queries.
         boto_client: botocore.client.BedrockRuntime
@@ -450,11 +481,14 @@ class EnergyWizardPostgres(EnergyWizardBase):
         self.psycopg2 = try_import('psycopg2')
 
         if meta_columns is None:
-            self.meta_columns = ['title', 'url', 'id']
+            self.meta_columns = self.DEFAULT_META_COLS
         else:
             self.meta_columns = meta_columns
 
-        assert 'id' in self.meta_columns, "Please include the 'id' column!"
+        assert 'id' in self.meta_columns, ("Please include the chunk id "
+                                           "column: 'id'!")
+        assert 'nrel_id' in self.meta_columns, ("Please include the document "
+                                                "id column: 'nrel_id'!")
 
         if cursor is None:
             db_user = os.getenv("EWIZ_DB_USER")
@@ -477,16 +511,17 @@ class EnergyWizardPostgres(EnergyWizardBase):
 
         if boto_client is None:
             access_key = os.getenv('AWS_ACCESS_KEY_ID')
-            secret_key = os.getenv('AWS_SECRET_ACCESS_KEY')
-            session_token = os.getenv('AWS_SESSION_TOKEN')
-            assert access_key is not None, "Must set AWS_ACCESS_KEY_ID!"
-            assert secret_key is not None, ("Must set AWS_SECRET_ACCESS_KEY!")
-            assert session_token is not None, "Must set AWS_SESSION_TOKEN!"
-            self._aws_client = boto3.client(service_name='bedrock-runtime',
-                                            region_name='us-west-2',
-                                            aws_access_key_id=access_key,
-                                            aws_secret_access_key=secret_key,
-                                            aws_session_token=session_token)
+            secret = os.getenv('AWS_SECRET_ACCESS_KEY')
+            session = os.getenv('AWS_SESSION_TOKEN')
+            if access_key and secret and session:
+                self._aws_client = boto3.client(service_name='bedrock-runtime',
+                                                region_name='us-west-2',
+                                                aws_access_key_id=access_key,
+                                                aws_secret_access_key=secret,
+                                                aws_session_token=session)
+            else:
+                self._aws_client = boto3.client(service_name='bedrock-runtime',
+                                                region_name='us-west-2')
         else:
             self._aws_client = boto_client
 
@@ -652,10 +687,13 @@ class EnergyWizardPostgres(EnergyWizardBase):
                                    "connection or query.")
 
         unique_ref_list = []
+        unique_nrel_ids = set()
         for ref_dict in ref_list:
-            if any(ref_dict == d for d in unique_ref_list):
+            if ref_dict['nrel_id'] in unique_nrel_ids:
                 continue
             unique_ref_list.append(ref_dict)
+            unique_nrel_ids.add(ref_dict['nrel_id'])
+
         ref_list = unique_ref_list
 
         if 'id' in ref_list[0]:
